/**The MIT License (MIT)
 uint8_t changeButtonScreen
 Copyright (c) 2018 by ThingPulse Ltd., https://thingpulse.com
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 */

// Modifications (from the ThingPulse original code) by http://www.frogymandias.org
// MIT license is maintained

/*****************************
 * Important: see settings.h to configure your settings!!!
 * ***************************/
#include "settings.h"

#include <Arduino.h>
#include "time.h"
#include <SPI.h>
#include <ESP8266WiFi.h>

// (Frogymansias mod)
#define VERSION "2022-11-27"

// (Frogymandias mod) Defined (or not) in settings.h
// No touch screen leaves more ESP8266 pins available for other uses
#ifdef TOUCH  // Also see #define HAVE_TOUCHPAD in settings.h
  #include <XPT2046_Touchscreen.h>
  #include "TouchControllerWS.h"
#endif

// (Frogymandias mod) See settings.h for more DHT settings
bool gotDHT = false;  // This is only set to be true after getting a valid DHT reading.
                      // So we can do a civilized display if people define GOTDHT but never
                      // actually onnect one.
#ifdef GOTDHT
  #include <Adafruit_Sensor.h>
  #include <DHT.h>
  #include <DHT_U.h>
#endif

// (Frogymandias mod)
#ifdef SHOWMOON
  #include "SunMoonCalc.h"
  #include "moonphases.h"
#endif

/***
   Install the following libraries through Arduino Library Manager
   - Mini Grafx by Daniel Eichhorn
   - ESP8266 WeatherStation by Daniel Eichhorn
   - Json Streaming Parser by Daniel Eichhorn
 ***/

#include <JsonListener.h>
#include <OpenWeatherMapCurrent.h>
#include <OpenWeatherMapForecast.h>
#include <MiniGrafx.h>
// (Frogymandias mod) 
#ifdef CAROUSEL
  #include <Carousel.h>
#endif
#include <ILI9341_SPI.h>

#include "ArialRounded.h"
// #include "moonphases.h"
#include "weathericons.h"

#define MINI_BLACK 0
#define MINI_WHITE 1
#define MINI_YELLOW 2
#define MINI_BLUE 3

#define MAX_FORECASTS 12

// defines the colors usable in the paletted 16 color frame buffer
uint16_t palette[] = {ILI9341_BLACK, // 0
                      ILI9341_WHITE, // 1
                      ILI9341_YELLOW, // 2
                      0x7E3C
                     }; //3

// Limited to 4 colors due to memory constraints
int BITS_PER_PIXEL = 2; // 2^2 =  4 colors

ADC_MODE(ADC_VCC);  // The single analog input is now dedicated to reading vcc. See ESP.getVcc() (below)
                    // See https://arduino-esp8266.readthedocs.io/en/latest/libraries.html#esp-specific-apis

ILI9341_SPI tft = ILI9341_SPI(TFT_CS, TFT_DC);
MiniGrafx gfx = MiniGrafx(&tft, BITS_PER_PIXEL, palette);

#ifdef CAROUSEL
  Carousel carousel(&gfx, 0, 0, 240, 100);
#endif 

// (Frogymandias mod) 
#ifdef TOUCH
  XPT2046_Touchscreen ts(TOUCH_CS, TOUCH_IRQ);
  TouchControllerWS touchController(&ts);
  
  void calibrationCallback(int16_t x, int16_t y);
  CalibrationCallback calibration = &calibrationCallback;
#endif

OpenWeatherMapCurrentData currentWeather;
OpenWeatherMapForecastData forecasts[MAX_FORECASTS];

// (Frogymandias mod) 
#ifdef SHOWMOON
SunMoonCalc::Moon moonData;
#endif

String feel;  // Do feel.reserve(nn) to reserve nn bytes (to reduce heap fragmentation.)

// TOC
// (Frogymandias mods)
float fixMetricMiles(float m);  // Convert meters to miles to fix Visibilty (that did not honor IS_METRIC)
float fixMetricRain(float mm);  // Ditto, for rain.  Was displaying forecasts for 20+ inches of rain.
String windDir (float dir);     // Converts wind direction from degrees to N, NSE, notation
void drawMillisRemaining();     // Display the amount of seconds until the next update.
float calcHeatIndex(float T, float RH); // US NWS formula for heat index
float calcWindChill(float T, float V);  // Wind chill
float calcDewpoint (float Tf, float RH);  // Magnus formula for dewpoint
String feelsLike(float TF, float humidity, bool shortDisplay);  // Converts dewpoint to text ("Getting muggy", etc.)
void shortDrawCurrentWeatherDetail(); // For displaying on screen 0 (taking the place of the carousel)

void initData();
void updateData();    // Get current and forecast weather data.  Moon data.  (Fogymandias) DHT temp/humidity data
void drawProgress(uint8_t percentage, String text); // Draw progress bars
void drawTime();        // Draw the time in big numbers at the top of screen 0.  And a few things more ear there. 
void drawWifiQuality(); // What it says.  In the upper right of screen 0.
void drawCurrentWeather();  // Draw current weather big icon and current temp in big numbers, below the time.
void drawForecast();        // Doesn't exist(?)
void drawForecastDetail(uint16_t x, uint16_t y, uint8_t dayIndex);  // Helper for displaying forecasts
void drawAstronomy(); // Sunrise, set, moon icon, etc. at the bottom of the screen
char determineMoonIcon();
void drawCurrentWeatherDetail();  // Screen 1.  In text mode using only body text, no icons.
void drawLabelValue(uint8_t line, String label, String value);  // Helper for drawCurrentWeatherDetail()
void drawForecastTable(uint8_t start);  // Screens 2 and 3. Four forcasts per screen.  Has icons.
void drawAbout(); // Display logo, technical information.  Free mem, various chip details, uptime, etc.
void drawSeparator(uint16_t y); // Does not exist.
String getTime(time_t *timestamp);  // Syncs to time server
const char* getMeteoconIconFromProgmem(String iconText);
const char* getMiniMeteoconIconFromProgmem(String iconText);

// (Frogymandias mod) 
#ifdef CAROUSEL
  void drawForecast1(MiniGrafx *display, CarouselState* state, int16_t x, int16_t y);
  void drawForecast2(MiniGrafx *display, CarouselState* state, int16_t x, int16_t y);
  void drawForecast3(MiniGrafx *display, CarouselState* state, int16_t x, int16_t y);
#endif

void loadPropertiesFromSpiffs();

// (Frogymandias mod) 
#ifdef CAROUSEL
  FrameCallback frames[] = { drawForecast1, drawForecast2, drawForecast3 };
  int frameCount = 3;
#endif

// (Frogymandias mod) 
#ifdef GOTDHT
  // create DHT22 instance (defined in settings.h)
  DHT_Unified dht(DHT_DATA_PIN, DHTTYPE);
  float dhtTempC = -1;
  float dhtTempF = -1;
  float dhtHumidity = -1;
  float oldTempC = -1;
  float oldTempF = -1;
  float oldHumidity = -1;
  bool badDhtTemp = false;
  bool badHumidity = false;
#else
  bool badDhtTemp = true;
  bool badHumidity = true;
#endif

bool tftLedOff = false; // (Frogymandias)
  
// (Frogymandias mod) These are used for both TOUCH and !TOUCH
// For new calcHeatIndex(), calcWindChill(), and calcDewpoint()
float heatIndex = -1;
float heatIndexF = -1;
float heatIndexC = -1;

const float TWCF = 50;                      // Wind chill threshold in F
                                            // (Wind chill only valid T <= 50F and WS > 3mph)
//const float TWCC = (TWCF-32) / .5556;       // Wind chill threshold in C
const float TWCC = 10;
float TWC = (IS_METRIC ? TWCC : TWCF);  // This gets changed when we toggle IS_METRIC

float dewPoint = -1;
float dewPointF = -1;
float dewPointC = -1;

float T = 400;
float TF = 400;
float TC = 400;
float windSpeed = 400;
float windSpeedF = 400;
float windSpeedC = 400;
float windChill = 400;
float windChillF = 400;
float windChillC = 400;

// Does not work here!!!
//String degreeSign = "°F";
//if (IS_METRIC) {
//  degreeSign = "°C";
//}

//String degreeSign = (IS_METRIC ? "°C" : °F");

// how many different screens do we have?
int screenCount = 5;
long lastDownloadUpdate = millis();

uint8_t screen = 0;
// divide screen into 4 quadrants "< top", "> bottom", " < middle "," > middle "
uint16_t dividerTop, dividerBottom, dividerMiddle;

// (Frogymandias mod) 
#ifdef TOUCH
  uint8_t changeScreen(TS_Point p, uint8_t screen);
#endif

// (Frogymandias mod) 
#ifdef BUTTON_SWITCHES
  uint8_t readButton1(uint8_t screen);
  uint8_t readButton2(uint8_t screen);
  uint8_t readButtons(uint8_t screen);
  uint8_t buttonState;
  uint8_t changeButtonScreen(uint8_t buttonState, uint8_t screen);
  bool buttonState1 = HIGH; // Assumed pull-up resisters for NO button switches
  bool buttonState2 = HIGH;

#endif

  //if (SLEEP_INTERVAL_SECS) { Sleep_Interval = true; }

long timerPress;
bool canBtnPress;
bool sleep_mode();  // Process sleep.  Am I asleep?  Should I wake?  Should I go to sleep?  etc.
char* make12_24(int hour);
bool isScreen0;

void connectWifi() {
  if (WiFi.status() == WL_CONNECTED) return;
  //Manual Wifi
  Serial.printf("Connecting to WiFi %s/%s", WIFI_SSID.c_str(), WIFI_PASS.c_str());
  WiFi.disconnect();
  WiFi.mode(WIFI_STA);
  WiFi.hostname(WIFI_HOSTNAME);
  WiFi.begin(WIFI_SSID.c_str(), WIFI_PASS.c_str());
  int i = 0;
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    if (i > 80) i = 0;
    drawProgress(i, "Connecting to WiFi '" + String(WIFI_SSID.c_str()) + "'");
    i += 10;
    Serial.print(".");
  }
  drawProgress(100, "Connected to WiFi '" + String(WIFI_SSID.c_str()) + "'");
  Serial.println("connected.");
  Serial.printf("Connected, IP address: %s/%s\n", WiFi.localIP().toString().c_str(), WiFi.subnetMask().toString().c_str()); //Get ip and subnet mask
  Serial.printf("Connected, MAC address: %s\n", WiFi.macAddress().c_str());  //Get the local mac address
}

void initTime() {
  time_t now;

  gfx.fillBuffer(MINI_BLACK);
  gfx.setFont(ArialRoundedMTBold_14);

  Serial.printf("Configuring time for timezone %s\n", TIMEZONE.c_str());
  configTime(TIMEZONE.c_str(), NTP_SERVERS);
  int i = 1;
  while ((now = time(nullptr)) < NTP_MIN_VALID_EPOCH) {
    drawProgress(i * 10, "Updating time...");
    Serial.print(".");
    delay(300);
    yield();
    i++;
  }
  drawProgress(100, "Time synchronized");
  Serial.println();

  printf("Local time: %s", asctime(localtime(&now))); // print formated local time, same as ctime(&now)
  printf("UTC time:   %s", asctime(gmtime(&now)));    // print formated GMT/UTC time
}

void setup() {
  Serial.begin(115200);

// (Frogymandias mod) 
#ifdef TOUCH
  loadPropertiesFromSpiffs();
#endif

// (Frogymandias mod) 
#ifdef BUTTON_SWITCHES
  pinMode(BUTTON_1, INPUT);
  pinMode(BUTTON_2, INPUT);
#endif

// (Frogymandias mod) 
#ifdef GOTDHT
  // wait for serial monitor to open
  //while(! Serial);
  //Serial.print("\n\n----DHT----\n\n");
  
  // initialize dht22
  dht.begin();
  delay(500);
#endif

  // The LED pin needs to set HIGH
  // Use this pin to save energy
  // Turn on the background LED
 #ifdef TFT_LED
  Serial.println(TFT_LED);
  pinMode(TFT_LED, OUTPUT);
  digitalWrite(TFT_LED, HIGH);    // HIGH to Turn on;
#endif
  gfx.init();
  gfx.fillBuffer(MINI_BLACK);
  gfx.commit();

// (Frogymandias mod) 
#ifdef TOUCH
  Serial.println("Initializing touch screen...");
  ts.begin();

  Serial.println("Mounting file system...");
  bool isFSMounted = SPIFFS.begin();
  if (!isFSMounted) {
    Serial.println("Formatting file system...");
    drawProgress(50, "Formatting file system");
    SPIFFS.format();
  }
  drawProgress(100, "Formatting done");

  /* Allow user to force a screen re-calibration  */
  gfx.fillBuffer(MINI_BLACK);
  gfx.drawString(120, 160, F("Press and hold\nto initiate touch screen\ncalibration"));
  gfx.commit();
  delay(4000);
  yield();
  boolean isCalibrationAvailable = touchController.loadCalibration();
  if(ts.touched()) {
    isCalibrationAvailable = false;
    gfx.fillBuffer(MINI_YELLOW);
    gfx.drawString(120, 160, F("Calibration initiated\nnow release screen"));
    gfx.commit();

    // Wait for release otherwise touch becomes first calibration point
    while(ts.touched()) { 
      delay(10);
      yield();
    }
    delay(100); // debounce
    touchController.getPoint(); // throw away last point
  }

  if (!isCalibrationAvailable) {
    Serial.println("Calibration data not available or force calibration initiated");
    touchController.startCalibration(&calibration);
    while (!touchController.isCalibrationFinished()) {
      gfx.fillBuffer(0);
      gfx.setColor(MINI_YELLOW);
      gfx.setTextAlignment(TEXT_ALIGN_CENTER);
      gfx.drawString(120, 160, "Please calibrate\ntouch screen by\ntouch point");
      touchController.continueCalibration();
      gfx.commit();
      yield();
    }
    touchController.saveCalibration();
  }
#else
  Serial.println(" (No touch screen)");
#endif

  dividerTop = 64;
  dividerBottom = gfx.getHeight() - dividerTop;
  dividerMiddle = gfx.getWidth() / 2;

  connectWifi();

// (Frogymandias mod) 
#ifdef CAROUSEL
  carousel.setFrames(frames, frameCount);
  carousel.disableAllIndicators();
#endif

  initTime();

  // update the weather information
  updateData();
  timerPress = millis();
  canBtnPress = true;
}
// End setup()

long lastDrew = 0;
bool btnClick;

// (Frogymandias mod) 
#ifdef TOUCH
  uint8_t MAX_TOUCHPOINTS = 10;
  TS_Point points[10];
  uint8_t currentTouchPoint = 0;
#endif

#ifdef BUTTON_SWITCHES
  uint8_t button1State = 0;
  uint8_t button2State = 0;
  long buttonTimeUsed = 0;
 #endif
 
void loop() {
  static bool asleep = false;	//  asleep used to stop screen change after touch for wake-up
  gfx.fillBuffer(MINI_BLACK);

// (Frogymandias mod) 
#ifdef TOUCH
  /* Break up the screen into 4 sections a touch in section:
   * - Top changes the time format (??? is bottom)
   * - Left back one page
   * - Right forward one page
   * - Bottom jump to page 0  (??? is top)
   */
  if (touchController.isTouched(500)) {
    TS_Point p = touchController.getPoint();
    timerPress = millis();
    if (!asleep) { 				// no need to update or change screens;
    	screen = changeScreen(p, screen);
    }
  } // isTouched()
#endif

// (Frogymandias mod) 
#ifdef BUTTON_SWITCHES  // *Instead* of TOUCH screen
  buttonTimeUsed = 0;
  uint32_t buttonTimeStart = millis();
  long buttonLoopCnt = 0;
  while (buttonTimeUsed <= 800) { // Need to delay for about 800 ms for screen 0 seconds to display cleanly
    uint8_t button2State = 0;
    uint8_t button1State = readButton1(screen);
    if (button1State == 0 ) { button2State = readButton2(screen); }

    if (6 == 9) { // Disabling Very Long state.  For now
      if (button1State == BUTTON_1_VERY_LONG) button1State = BUTTON_1_LONG;
      if (button2State == BUTTON_2_VERY_LONG) button2State = BUTTON_2_LONG;
    }
    
    uint8_t buttonState = button1State | button2State;  
    if (buttonState > 0) { timerPress = millis(); }
    // We don't need to turn the screen off now that Long Press that non-TOUCH can do sleep.  Clean this out.
    // Or not.  Toggling LED on/off with Very Long Press is manually initiiated sleep.  Maybe should keep.
    if ( ((button1State == BUTTON_1_VERY_LONG) || (button2State == BUTTON_2_VERY_LONG)) && 6 == 6) { // Enabling Very Long, for now
      Serial.printf("We had a very long press.  turning LED off. Wait....\n"); // AKA manually initiiated sleep
        int s = 0;
        do {
          drawProgress(s, "Turning off the display");
          delay(10/10);
          yield();
        } while (s++ < 100 && !readButtons(screen) );  // 100
        if ( !(s < 100) ) { 
          digitalWrite(TFT_LED, LOW);
          tftLedOff = true;
          Serial.printf("OK, the LED is off.  The buttons should respond now.\n");
        } else {
          Serial.printf("Pressed a key so not going to sleep\n");
        }
      
    } else if ( (button1State == BUTTON_1_LONG)) {  // Maybe should do a bitwise test instead of a == test.
        if (tftLedOff == true) {
          digitalWrite(TFT_LED, HIGH);
          tftLedOff = false;
          Serial.printf("(Longpress) Turning LED back on.\n");
        } else {
#ifdef TOGGLE_METRIC_1
        if (isScreen0) {      
          IS_METRIC=!IS_METRIC;
        //initData();     //  Update current (and not forecasts) data
          updateData();   // Get current and forecasts
          Serial.printf("Toggling IS_METRIC\n");
        }
#endif 
      screen = 0;
      isScreen0 = true;
      Serial.printf("We had a long press.  Setting screen to 0\n"); 
      }
    } else if (button2State == BUTTON_2_LONG) { 
      if (tftLedOff == true) {
          digitalWrite(TFT_LED, HIGH);
          tftLedOff = false;
          Serial.printf("(Longpress) Turning LED back on.\n");
      } else if (SLEEP_INTERVAL_SECS && screen == 0 && TOGGLE_SLEEP) {
        // Toggle sleep state
        Sleep_Interval = !Sleep_Interval;
        Serial.printf("%s sleep mode.\n", (Sleep_Interval ? "Enabling" : "Disabling"));
      } else if (TOGGLE_12HR && screen == 0) {
        IS_STYLE_12HR = !IS_STYLE_12HR;
        Serial.printf("%s mode", (IS_STYLE_12HR) ? "AM/PM" : "24 hour");
      } else {
        //Serial.printf("BUTTON_2_LONG isn't defined to do anything\n");
  
#ifdef TOGGLE_METRIC_2  
        if (isScreen0) {      
          IS_METRIC=!IS_METRIC;
          initData();     //  Update current (and not forecasts) data
          //updateData(); // Get current and forecasts
          Serial.printf("Toggling IS_METRIC\n");
        }
#endif        
        Serial.printf("We had a long press.  Setting screen to 0\n"); 
        screen = 0;
        isScreen0 = true;
        
      }
    } else if (buttonState > 0) {   // Short press
        if (tftLedOff == true) {
          digitalWrite(TFT_LED, HIGH);
          tftLedOff = false;
          Serial.printf("(Shortpress) Turning LED back on.\n");
        } else {
          screen = changeButtonScreen(buttonState, screen);
        } 
    }
    //delay(200); Serial.printf("buttonTimeUsed: %d\n",buttonTimeUsed);
    buttonLoopCnt++;
    buttonTimeUsed = buttonTimeUsed + (millis() - buttonTimeStart);
  } // while (buttonTimeUsed <= 800) {
  //Serial.printf("buttonLoopCnt = %d\n", buttonLoopCnt); // Varies betwen 400s to 500s
#else // Not doing CAROUSEL nor BUTTON_SWITCHES but we need to delay so the seconds will display evenly.
#ifndef TOUCH
    delay(800); // Approximation so the seconds display smoothly when the carousel isn't displayed.
#endif
#endif  // #ifdef BUTTON_SWITCHES

  if (!(asleep = sleep_mode())) {
    if (screen == 0) {
      drawTime();
      isScreen0 = true; // Affects about screen (Frogymandias mod)
      drawWifiQuality();
      drawMillisRemaining();
#ifdef CAROUSEL // (Frogymandias mod) 
      int remainingTimeBudget = carousel.update();
      if (remainingTimeBudget > 0) {
        // You can do some work here
        // Don't do stuff if you are below your
        // time budget.
        //Serial.println("remainingTimeBudget: " + String(remainingTimeBudget,0));
        delay(remainingTimeBudget);
      }
#endif
      drawCurrentWeather(); // screen == 0
// (Frogymandias mod) 
#ifdef SHOWMOON
      drawAstronomy();
#endif

    } else if (screen == 1) {
      isScreen0 = false;
      drawCurrentWeatherDetail();
    } else if (screen == 2) {
      isScreen0 = false;
      drawForecastTable(0);
    } else if (screen == 3) {
      isScreen0 = false;
      drawForecastTable(4);
    } else if (screen == 4) {
      isScreen0 = false;
      drawAbout();
    }
    gfx.commit();

    // Check if we should update weather information
    if (millis() - lastDownloadUpdate > 1000 * UPDATE_INTERVAL_SECS) {
      updateData();
      lastDownloadUpdate = millis();
    }
  } //!asleep
} // loop()

  /*
    	Check and activate when it is time to go to sleep
  
    	parameters:	(defined in settings)
    	 	SLEEP_INTERVAL_SECS 	time between screen touches in seconds before activating sleep mode
    	 	HARD_SLEEP			 	true  -> deep sleep requiring interrupt or reset to wake
    	 							false -> soft sleep turning off backlight wake with screen press
  
    	returns: true when sleep mode is active
  */
  bool sleep_mode() {
    static bool sleeping = false; // no need to waste time painting going to sleep screens
    // Sleep_Interval is a Frogymandias mod (it enables sleep mode so we can enable/disable sleeping on the fly.)
    if ((SLEEP_INTERVAL_SECS && Sleep_Interval) 
        && millis() - timerPress >= SLEEP_INTERVAL_SECS * 1000) { // after SLEEP_INTERVAL_SECS go to sleep
      if (true == sleeping)
        return sleeping;  // all-ready asleep

      Serial.printf("Going to Sleep! Screen %d\n", screen);
      int s = 0;
      do {
        drawProgress(s, "Going to Sleep!");
        delay(10);
        yield();
 #ifdef TOUCH
      } while (s++ < 100 && !touchController.isTouched(0));
      if (s < 100) {                         // early exit abort
        timerPress = millis();               // reset sleep timeout
        touchController.getPoint();          // throw away
        if (touchController.isTouched(0))    // resets lastTouched
          touchController.getPoint();        // throw away
#endif  // #ifdef TOUCH

#ifdef BUTTON_SWITCHES
      } while (s++ < 100 && !readButtons(0));
      if (s < 100) {                         // early exit abort
        timerPress = millis();               // reset sleep timeout
          if (readButtons(0)) {   // resets lastTouched
          //int t = readButtons(0)
          delay(200);
          Serial.printf("Not going to sleep. Screen is %d\n", screen);
        }
#else
         //} while (s++ < 100);
          //  if (s < 100) {                         // early exit abort
          //      //delay(20);
           // }
#endif // #ifdef BUTTON_SWITCHES
      } else {
        Serial.printf("Gone to sleep. Screen %d\n", screen);
        delay(200);
        sleeping = true;
        if (true == HARD_SLEEP) {
          // go to deepsleep for xx minutes or 0 = permanently
          ESP.deepSleep(0, WAKE_RF_DEFAULT); // 0 delay = permanently to sleep
        } else {
#ifdef TFT_LED
          digitalWrite(TFT_LED, LOW);        // Back light OFF
          tftLedOff = true;
          Serial.printf("Slept. Back light is off. Screen %d\n", screen);
#endif
        }
      }
    } else {                                 // Not time to sleep
      if (sleeping) {                        // Wake up
#ifdef TFT_LED
        tftLedOff = false;
        digitalWrite(TFT_LED, HIGH);         // Back light ON
#ifdef BUTTON_SWITCHES
        //screen=0; // Cringe.  Force screen to 0 after button mode wakeup. Fixed.
#endif
        Serial.printf("(Waking from sleep.) Back light is on. Screen is %d\n", screen);
#endif  // #ifdef TFT_LED
        sleeping = false;
      }
    }
    return sleeping;  // used to prevent screen changes on wake-up screen press
  }	// sleep_mode()

void initData() {
  OpenWeatherMapCurrent *currentWeatherClient = new OpenWeatherMapCurrent();
  currentWeatherClient->setMetric(IS_METRIC);
  currentWeatherClient->setLanguage(OPEN_WEATHER_MAP_LANGUAGE);
  currentWeatherClient->updateCurrentById(&currentWeather, OPEN_WEATHER_MAP_API_KEY, OPEN_WEATHER_MAP_LOCATION_ID);

// (Frogymandias mods) 
  // Calced using the official US NWS formula

//  TWCF = 50;                      // Wind chill threshold in F
                                  // (Wind chill only valid T <= 50F and WS > 3mph)
//const float TWCC = (TWCF-32) / .5556; // Wind chill threshold in C
//  TWCC = 10;
  TWC = (IS_METRIC ? TWCC : TWCF);
 
  T = currentWeather.temp;  // (Already has had IS_METRIC applied.)
  TF = (IS_METRIC ? (T * 1.8) + 32 : T ); // Is T already F?  If not, convert
  TC = (IS_METRIC ? T : (T - 32) * .5556);  // Is T already C?  If not, convert

  heatIndexF = calcHeatIndex(TF, currentWeather.humidity);  // Ignores IS_METRIC; must be in F
  heatIndexC = (heatIndexF - 32) * .5556;
  heatIndex = (IS_METRIC ? heatIndexC : heatIndexF);
  
  windSpeed = currentWeather.windSpeed; // (Already has had IS_METRIC applied.) 
  windSpeedC = (IS_METRIC ? windSpeed : windSpeed / 2.237); // m/s (meters per second)
  windSpeedF = (IS_METRIC ? windSpeed * 2.237 : windSpeed); // mph (miles per hour)
  
  windChillF = calcWindChill(TF, windSpeedF); // Wind Chill must be forced to F
  windChillC = (windChillF - 32) * .5556;
  windChill = (IS_METRIC ? windChillC : windChillF );
  
  dewPointF = calcDewpoint(TF, currentWeather.humidity); // calcDewpoint() assumes T is F, returns Tdp as F)
  dewPointC = (dewPointF - 32) * .5556;
  dewPoint = (IS_METRIC ? dewPointC : dewPointF);

#ifdef VERBOSE_CONSOLE
  Serial.println("------------------------------");
  Serial.println("T: " + String(T,1) + " windSpeed: " + String(windSpeed,1) + " windChill: " + String(windChill,1)+ " heatIndex: " + heatIndex );
  Serial.println("TempF: " + String(TF,1) + "°F WSF: " + String(windSpeedF,1) + "mph"+ " WindchillF: " + String(windChillF,1) + "°F heatIndexF: " + heatIndexF);
  Serial.println("TempC: " + String(TC,1) + "°C WSC: " + String(windSpeedC,1) + "m/s"+ " WindchillC: " + String(windChillC,1) + "°C heatIndexC: " + heatIndexC);
  Serial.println("TWC: " + String(TWC,1) + " TWCF: " + String(TWCF,1) + + "F TWCC: " + String(TWCC,1) + "C");
  Serial.println("dewPoint:  " + String(dewPoint,1) + " dewPointF:  " + String(dewPointF,1) +"F dewPointC:  " + String(dewPointC,1) + "C");  
  Serial.println("------------------------------");
#endif  
  delete currentWeatherClient;
  currentWeatherClient = nullptr;


}

// Update the internet based information and update screen
void updateData() {
  time_t now = time(nullptr);

  gfx.fillBuffer(MINI_BLACK);
  gfx.setFont(ArialRoundedMTBold_14);

  drawProgress(50, "Updating conditions...");
  
  initData();

  drawProgress(70, "Updating forecasts...");
  OpenWeatherMapForecast *forecastClient = new OpenWeatherMapForecast();
  forecastClient->setMetric(IS_METRIC);
  forecastClient->setLanguage(OPEN_WEATHER_MAP_LANGUAGE);
  uint8_t allowedHours[] = {12, 0};
  //uint8_t allowedHours[] = {18, 12, 6, 0};
  //uint8_t allowedHours[] = {20, 16, 8, 4};
  
  forecastClient->setAllowedHours(allowedHours, sizeof(allowedHours));
  //forecastClient->setAllowedHours(allowedHours, 4);
  forecastClient->updateForecastsById(forecasts, OPEN_WEATHER_MAP_API_KEY, OPEN_WEATHER_MAP_LOCATION_ID, MAX_FORECASTS);
  delete forecastClient;
  forecastClient = nullptr;


// (Frogymandias mod)
#ifdef SHOWMOON
  drawProgress(80, "Updating astronomy...");
  // 'now' has to be epoch instant, lat/lng in degrees not radians
  SunMoonCalc *smCalc = new SunMoonCalc(now, currentWeather.lat, currentWeather.lon);
  moonData = smCalc->calculateSunAndMoonData().moon;
  delete smCalc;
  smCalc = nullptr;
#endif

// (Frogymandias mod)
#ifdef GOTDHT
  drawProgress(70, "Updating ambient temp...");
    // Get the data from the DHT sensor
  dhtTempC = oldTempC;
  dhtTempF = oldTempF;
  dhtHumidity = oldHumidity;
  
  sensors_event_t event;
  dht.temperature().getEvent(&event);
  if (!(isnan(event.temperature))) {
    dhtTempC = event.temperature;
    dhtTempF = (dhtTempC * 1.8) + 32;
    badDhtTemp = false;
    gotDHT = true;  
  } else {
    badDhtTemp = true;
  }
  oldTempC = dhtTempC;
  oldTempF = dhtTempF;
  
  dht.humidity().getEvent(&event);
  if (!(isnan(event.relative_humidity))) {
      dhtHumidity = event.relative_humidity;
      gotDHT = true;
      badHumidity = false;
  } else {
    badHumidity = true;
  }
  oldHumidity = dhtHumidity;
  if (gotDHT) {
    Serial.print(dhtTempF);
    Serial.print("F ");
    Serial.print(dhtHumidity);
    Serial.println("%");
  } else {
    Serial.println("We don't have a DHT temp/humidity device connected.");
  }
#endif // #ifdef GOTDHT

  Serial.printf("Free mem: %d\n",  ESP.getFreeHeap());

  delay(100); // Was 1000
}

// Progress bar helper
void drawProgress(uint8_t percentage, String text) {
  gfx.fillBuffer(MINI_BLACK);
  gfx.drawPalettedBitmapFromPgm(20, 5, ThingPulseLogo);
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  gfx.setColor(MINI_WHITE);
  gfx.drawString(120, 90, "https://thingpulse.com");
  //gfx.drawString(120, 110, "http://www.frogymandias.org");
  gfx.setColor(MINI_YELLOW);

  gfx.drawString(120, 146, text);
  gfx.setColor(MINI_WHITE);
  gfx.drawRect(10, 168, 240 - 20, 15);
  gfx.setColor(MINI_BLUE);
  gfx.fillRect(12, 170, 216 * percentage / 100, 11);

  gfx.commit();
}

// draws the clock
void drawTime() {
  char time_str[11];
  time_t now = time(nullptr);
  struct tm * timeinfo = localtime(&now);

  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setColor(MINI_WHITE);
  String date = WDAY_NAMES[timeinfo->tm_wday] + " " + MONTH_NAMES[timeinfo->tm_mon] + " " + String(timeinfo->tm_mday) + " " + String(1900 + timeinfo->tm_year);
  gfx.drawString(123, 6, date); // 120,6

  gfx.setFont(ArialRoundedMTBold_36);

  if (IS_STYLE_12HR) {                                                              //12:00
    int hour = (timeinfo->tm_hour + 11) % 12 + 1; // take care of noon and midnight
    if (IS_STYLE_HHMM) {
      sprintf(time_str, "%2d:%02d\n", hour, timeinfo->tm_min);                //hh:mm
    } else {
      sprintf(time_str, "%2d:%02d:%02d\n", hour, timeinfo->tm_min, timeinfo->tm_sec); //hh:mm:ss
    }
  } else {                                                                            //24:00
    if (IS_STYLE_HHMM) {
        sprintf(time_str, "%02d:%02d\n", timeinfo->tm_hour, timeinfo->tm_min); //hh:mm
    } else {
        sprintf(time_str, "%02d:%02d:%02d\n", timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec); //hh:mm:ss
    }
  }
  gfx.drawString(120, 20, time_str);

  gfx.setTextAlignment(TEXT_ALIGN_LEFT);
  gfx.setFont(ArialMT_Plain_10);
  gfx.setColor(MINI_BLUE);
  if (IS_STYLE_12HR) {
    sprintf(time_str, "\n%s", timeinfo->tm_hour >= 12 ? "PM" : "AM");
    gfx.drawString(195, 27, time_str);
  }
}

// (Frogymandias mod)
// Draw seconds remaining until the next update
void drawMillisRemaining() {
    int y=6;      // Let's be parametric
    int yInc=18;  // Space between lines
    gfx.setTransparentColor(MINI_BLACK);
    gfx.setTextAlignment(TEXT_ALIGN_LEFT);
    //gfx.setFont(ArialMT_Plain_10);
    gfx.setFont(ArialRoundedMTBold_14);
    gfx.setColor(MINI_BLUE);

    char remainingBuf[22];
    int remainingSecs = UPDATE_INTERVAL_SECS-(millis() - lastDownloadUpdate)/1000;
 //#ifdef CAROUSEL  // Drawing remaining seconds all the time louses up the carousel timing
 #ifndef NEVER  // Let's always only draw the last 30 secs before update
    if (remainingSecs <= (30) ) { // So we will only do it for the last 30 seconds
          gfx.setColor(MINI_WHITE);
         sprintf(remainingBuf,"Rm: %d",remainingSecs);  // int to string
         gfx.drawString(1,y,remainingBuf); y=y+yInc;
         //gfx.setFont(ArialMT_Plain_10);
         gfx.drawString(1,y,SHORT_LOCATION_NAME); y=y+yInc;
    } else {
         gfx.drawString(1,y,SHORT_LOCATION_NAME); y=y+yInc;
    }
#else // We aren't using the carousel so we aren't worrying about its timing.
    if (remainingSecs <= (30) ) { // We will change the color for the last few seconds.  For warning.
          gfx.setColor(MINI_WHITE);
    }
   sprintf(remainingBuf,"Rm: %d",remainingSecs);  // int to string
   gfx.drawString(1,y,remainingBuf); y=y+yInc;
   //gfx.setFont(ArialMT_Plain_10);
   gfx.drawString(1,y,SHORT_LOCATION_NAME); y=y+yInc;
#endif

    // Sleep_Interval is a Frogymandias mod
     if (Sleep_Interval && SLEEP_INTERVAL_SECS) {
       gfx.setFont(ArialRoundedMTBold_14);
       int remainingSleep = SLEEP_INTERVAL_SECS-((millis() - timerPress)/1000);
       if (remainingSleep >= 15) {
        sprintf(remainingBuf,"Sleep");       
       } else {
        sprintf(remainingBuf,"Sl: %d",remainingSleep);  // int to string
       }
       gfx.drawString(1,y,remainingBuf); y=y+yInc;
       gfx.drawString(1,y,"mode"); y = y+yInc;
     }
}

// draws current weather information (And some forecast information)
void drawCurrentWeather() {

// (Frogymandias mod)
  //float T = currentWeather.temp;
  //float V = currentWeather.windSpeed;
  float THI = (IS_METRIC ? 27 : 78);  // Heat index threshold Usually around 80F. (27C) 

  //float TWC = (IS_METRIC ? TWCC : TWCF);   // Wind chill threshold Usually around 32F. (0C)
  
  gfx.setTransparentColor(MINI_BLACK);
  gfx.drawPalettedBitmapFromPgm(0, 55, getMeteoconIconFromProgmem(currentWeather.icon));
  // Weather Text

  gfx.setFont(ArialRoundedMTBold_14);

// (Frogymandias mod)
#ifdef NEVER // We now display a shortened  version of this elsewhere
  gfx.setColor(MINI_BLUE);
  gfx.setTextAlignment(TEXT_ALIGN_RIGHT);
  gfx.drawString(220, 60, DISPLAYED_LOCATION_NAME); // y = 65
#endif

// (Frogymandias mod)
#ifndef TOUCH
// We draw two mini-forcasts where the outside temp used to be
// Derived from drawForecastTable()
  //uint8_t i = 0;
  int firstColumnLabelX = 90;
  int firstColumnValueX = 110;
  int y = 59;
  int yInc = 15;
  
 String degreeSign = "°F";
  if (IS_METRIC) {
   degreeSign = "°C";
  }
  for (uint8_t i = 0; i < 2; i++) {
    gfx.setColor(MINI_YELLOW);
    gfx.setTextAlignment(TEXT_ALIGN_LEFT);
    time_t time = forecasts[i].observationTime;
    struct tm * timeinfo = localtime (&time);
    if (IS_STYLE_12HR) {
      gfx.drawString(firstColumnLabelX, y, WDAY_NAMES[timeinfo->tm_wday] + " " + String(make12_24(timeinfo->tm_hour)));
    } else {
      gfx.drawString(firstColumnValueX, y, WDAY_NAMES[timeinfo->tm_wday] + " " + String(timeinfo->tm_hour) + ":00");
    }
    gfx.setColor(MINI_BLUE);
    gfx.drawString(firstColumnLabelX, y + yInc*1, "T:");
    gfx.setColor(MINI_WHITE);
    gfx.drawString(firstColumnValueX, y + yInc*1, String(forecasts[i].temp, 0) + degreeSign);

    if (forecasts[i].rain >= .1) {  // Will there be any percipitation?
      gfx.setColor(MINI_BLUE);      // Yes.  Then this is what we want to display
      gfx.drawString(firstColumnLabelX, y + yInc*2, "P: ");
      gfx.setColor(MINI_WHITE);
      gfx.drawString(firstColumnValueX, y + yInc*2, String(fixMetricRain(forecasts[i].rain), 2) + (IS_METRIC ? "mm" : "in"));
    } else {  // No percipitation, so display humidity
      gfx.setColor(MINI_BLUE);
      gfx.drawString(firstColumnLabelX, y + yInc*2, "H: ");
      gfx.setColor(MINI_WHITE);  
      gfx.drawString(firstColumnValueX, y + yInc*2, String(forecasts[i].humidity) + "%");
    }
    gfx.setColor(MINI_BLUE);
    gfx.drawString(firstColumnLabelX, y + yInc*3, "W: ");
    gfx.setColor(MINI_WHITE);  
    gfx.drawString(firstColumnValueX, y + yInc*3, String(forecasts[i].windSpeed, 0) + (IS_METRIC ? "m/s" : "mph"));

    firstColumnLabelX = firstColumnLabelX + 78;
    firstColumnValueX = firstColumnValueX + 78;
 } // for (uint8_t i = 0; ...
 #endif

 // (Frogymandias mod)
#ifdef CAROUSEL
  gfx.setFont(ArialRoundedMTBold_36);
  gfx.setColor(MINI_WHITE);
  gfx.setTextAlignment(TEXT_ALIGN_RIGHT);
  gfx.drawString(220, 78, String(currentWeather.temp, 1) + (IS_METRIC ? "°C" : "°F"));
  //  gfx.drawString(220, 60, DISPLAYED_LOCATION_NAME); // y = 65
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setColor(MINI_YELLOW);
  gfx.setTextAlignment(TEXT_ALIGN_RIGHT);
 if (T >= THI && heatIndex > T) { // Normally is &&
    gfx.drawString(220, 60, "H: " + String(currentWeather.humidity) + "%");
    gfx.drawString(163, 60, "HI: " +String(heatIndex, 1) + (IS_METRIC ? "°C" : "°F"));
    // Need to add chill factor
 } else if (T <= TWC && T > windChill) {
    gfx.drawString(220, 60, "H: " + String(currentWeather.humidity) + "%");
    gfx.drawString(163, 60, "WC: " +String(windChill, 1) + (IS_METRIC ? "°C" : "°F"));
 } else {
    gfx.drawString(220, 60, "Humidity: " + String(currentWeather.humidity) + "%");
 }
#endif
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setColor(MINI_YELLOW);
  gfx.setTextAlignment(TEXT_ALIGN_RIGHT);
  gfx.drawString(220, 118, currentWeather.description);

// (Frogymandias mod)
#ifndef CAROUSEL
// Also could test gotDHT.  Maybe
  // We put temp(s) and humidity(s) where the carousel used to be.
  //gfx.setFont(ArialRoundedMTBold_14);
  gfx.setFont(ArialMT_Plain_10);
  gfx.setColor(MINI_BLUE);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  uint8_t line = 10;  // line height
  gfx.drawString(10, 141, "O");
  gfx.drawString(10, 141+line*1, "U");
  gfx.drawString(10, 141+line*2, "T");

// (Frogymandias mod)
#ifdef GOTDHT
  if (gotDHT) {
    gfx.drawString(10, 185+15, "I");
    gfx.drawString(10, 185+15+line*1, "N");
  }
#endif
      
  gfx.setFont(ArialRoundedMTBold_36);
  gfx.setColor(MINI_WHITE);
  gfx.setTextAlignment(TEXT_ALIGN_RIGHT);
  gfx.drawString(130, 135, String(currentWeather.temp, 1) + (IS_METRIC ? "°C" : "°F"));
  //  drawLabelValue(3, "Humidity:", String(currentWeather.humidity) + "%");
  gfx.drawString(220, 135, String(currentWeather.humidity) + "%");
#endif

// (Frogymandias mod) Pretty much assumes that carousel isn't being done.
#ifdef GOTDHT
  //Serial.print(dhtTempF);
  //Serial.print("F ");
  //Serial.print(dhtHumidity);
  //Serial.println("%");
  
  // DHT22 (or other) data goes here
  if (gotDHT) {
    gfx.setColor((badDhtTemp ? MINI_BLUE : MINI_WHITE));
    gfx.drawString(130, 170+15, String((IS_METRIC ? dhtTempC : dhtTempF), 1) + (IS_METRIC ? "°C" : "°F"));
    gfx.setColor((badHumidity ? MINI_BLUE : MINI_WHITE));
    gfx.drawString(220, 170+15, String(dhtHumidity,0) + "%");
  }  
  isScreen0 = true;
  shortDrawCurrentWeatherDetail();
#endif
}

// Enabling/disabling carousel is a Frogymandias mod
#ifdef CAROUSEL
  void drawForecast1(MiniGrafx *display, CarouselState* state, int16_t x, int16_t y) {
    drawForecastDetail(x + 10, y + 165, 0);
    drawForecastDetail(x + 95, y + 165, 1);
    drawForecastDetail(x + 180, y + 165, 2);
  }
  
  void drawForecast2(MiniGrafx *display, CarouselState* state, int16_t x, int16_t y) {
    drawForecastDetail(x + 10, y + 165, 3);
    drawForecastDetail(x + 95, y + 165, 4);
    drawForecastDetail(x + 180, y + 165, 5);
  }
  
  void drawForecast3(MiniGrafx *display, CarouselState* state, int16_t x, int16_t y) {
    drawForecastDetail(x + 10, y + 165, 6);
    drawForecastDetail(x + 95, y + 165, 7);
    drawForecastDetail(x + 180, y + 165, 8);
  }
  
  // helper for the forecast columns
  void drawForecastDetail(uint16_t x, uint16_t y, uint8_t dayIndex) {
    gfx.setColor(MINI_YELLOW);
    gfx.setFont(ArialRoundedMTBold_14);
    gfx.setTextAlignment(TEXT_ALIGN_CENTER);
    time_t time = forecasts[dayIndex].observationTime;
    struct tm * timeinfo = localtime (&time);
      //  Added 24hr / 12hr conversion  // 
      if(IS_STYLE_12HR){
      gfx.drawString(x + 25, y - 15, WDAY_NAMES[timeinfo->tm_wday] + " " + String(make12_24(timeinfo->tm_hour)));
    } else {
      gfx.drawString(x + 25, y - 15, WDAY_NAMES[timeinfo->tm_wday] + " " + String(timeinfo->tm_hour) + ":00");
    }
  
    gfx.setColor(MINI_WHITE);
    gfx.drawString(x + 25, y, String(forecasts[dayIndex].temp, 1) + (IS_METRIC ? "°C" : "°F"));
  
    gfx.drawPalettedBitmapFromPgm(x, y + 15, getMiniMeteoconIconFromProgmem(forecasts[dayIndex].icon));
    gfx.setColor(MINI_BLUE);
    if (forecasts[dayIndex].rain > 0.05) {
      gfx.drawString(x + 25, y + 60, String(fixMetricRain(forecasts[dayIndex].rain), 1) + (IS_METRIC ? "mm" : "in"));
    } else {
     gfx.drawString(x + 25, y + 60, "H: " + String(currentWeather.humidity) + "%");
    }
  }
#endif // #ifdef CAROUSEL

// (Frogymandias mod)
#ifdef SHOWMOON 
// draw moonphase, sunrise/set and moonrise/set
void drawAstronomy() {

// (Frogymandias mod)
// So we can move the whole thing up or down
#ifdef CAROUSEL
  int titlesY = 259;  // Orig 259 (270 is as low as we can go)
  int padding = 2;     // Spread the lines apart some
#else
  int titlesY = 270; 
  int padding = 0;
#endif
  
  gfx.setFont(MoonPhases_Regular_36);
  gfx.setColor(MINI_WHITE);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  // gfx.drawString(120, titlesY + 15, String((char) (97 + (moonData.illumination * 26))));
  gfx.drawString(128, titlesY + 15 + padding, String(determineMoonIcon()));  // x=120

  gfx.setColor(MINI_WHITE);
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  gfx.setColor(MINI_YELLOW);
  gfx.drawString(120, titlesY, MOON_PHASES[moonData.phase.index]);

  gfx.setTextAlignment(TEXT_ALIGN_LEFT);
  gfx.setColor(MINI_YELLOW);
  gfx.drawString(5, titlesY, SUN_MOON_TEXT[0]);
  gfx.setColor(MINI_WHITE);
  time_t time = currentWeather.sunrise;
  gfx.drawString(5, titlesY + 16 + padding, SUN_MOON_TEXT[1] + ":");
  gfx.drawString(45, titlesY + 16 + padding, getTime(&time));
  time = currentWeather.sunset;
  gfx.drawString(5, titlesY + 31 + padding, SUN_MOON_TEXT[2] + ":");
  gfx.drawString(45, titlesY + 31 + padding, getTime(&time));

  gfx.setTextAlignment(TEXT_ALIGN_RIGHT);
  gfx.setColor(MINI_YELLOW);
  gfx.drawString(235, titlesY, SUN_MOON_TEXT[3]);
  gfx.setColor(MINI_WHITE);

  float lunarMonth = 29.53;
  // approximate moon age
  gfx.drawString(190, titlesY + 16 + padding, SUN_MOON_TEXT[4] + ":");
  gfx.drawString(235, titlesY + 16 + padding, String(moonData.age, 1) + "d");
  gfx.drawString(190, titlesY + 31 + padding, SUN_MOON_TEXT[5] + ":");
  gfx.drawString(235, titlesY + 31 + padding, String(moonData.illumination * 100, 0) + "%");
}

// The Moon Phases font has 26 icons for gradiations, 1 full icon, and 1 empty icon: https://www.dafont.com/moon-phases.font
// All of them are an approximation. 
// Depending on date and location they would have to be rotated left or right by a varying degree.
// GOTCHA  I: as we use white to display the moon icon, what is black on that font page (link above) will effectively be rendered white!
// GOTCHA II: illumination in the range {0,1} will with near certainty never be exactly 0 or 1; rounding is, therefore, essential to ever get full/new moon!
char determineMoonIcon() {
  char moonIcon;
  // index in range of 0..14
  char index = round(moonData.illumination * 14);
//  Serial.printf("Moon illumination: %f -> moon icon index: %d\n", moonData.illumination, index);
  if (moonData.phase.index > 4) {
    // waning (4 = full moon)
    moonIcon = currentWeather.lat > 0 ? MOON_ICONS_NORTH_WANING[index] : MOON_ICONS_SOUTH_WANING[index];
  } else {
    // waxing
    moonIcon = currentWeather.lat > 0 ? MOON_ICONS_NORTH_WAXING[index] : MOON_ICONS_SOUTH_WAXING[index];
  }
  return moonIcon;
}
#endif  // SHOWMOON

// Wild guess because we don't know what is going on with Imperial rain amounts.
// Without the correction we are getting rain predictions close to 20 inches!
float fixMetricRain(float mm) {
  return (IS_METRIC ? mm : mm/25.4);  // Presumed to really be millimeters regardless of IS_METRIC setting.  ??
}

// Current visability isn't honoring IS_METRIC
// We are converting meters to miles.  Imperial people prefer visibility expressed as miles
float fixMetricMiles(float m) {
  return (IS_METRIC ? m : (m * 3.280839895) / 5280);
}

// Assumes that we are not on screen 0 and can use the whole screen.
void drawCurrentWeatherDetail() {
  int y = 2;
  int i = 0;
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  gfx.setColor(MINI_WHITE);
  gfx.drawString(120, y, "Current Conditions"); i++; i++;
  // (Frogymandias mod)
#ifdef GOTDHT
      if (badDhtTemp == false || badHumidity == false) { 
      gfx.setColor(MINI_BLUE);
      gfx.drawString(120, y + i*15, "Inside");i++;
      if (badDhtTemp == false) {
        float inTemp = (IS_METRIC ? dhtTempC : dhtTempF );
        drawLabelValue(i, "Temperature:", String(inTemp, 1) + (IS_METRIC ? "°C" : "°F")); i++;
      }
      if ( badHumidity == false) {
        drawLabelValue(i, "Humidity:", String(dhtHumidity, 0) + "%"); i++;
      }
      if (badDhtTemp == false && badHumidity == false) {
        float TdpF = calcDewpoint(dhtTempF, dhtHumidity); // calcDewpoint() always returns F
        float TdpC = (TdpF - 32) / 1.8;
        drawLabelValue(i, "Dew Point", String( (IS_METRIC ? TdpC : TdpF), 1) + (IS_METRIC ? "°C" : "°F") ); i++;
        drawLabelValue(i, "Feels like it is", feelsLike(dhtTempF, dhtHumidity, true) ); i++; 
      }
      
      i++;
      gfx.setTextAlignment(TEXT_ALIGN_CENTER);
      gfx.setColor(MINI_BLUE);
      gfx.drawString(120, y + i*15, "Outside");i++;
    }
#endif
    
    //float T = currentWeather.temp;
    //float TF = (IS_METRIC ? (T * 1.8) + 32 : T);
    //float V = currentWeather.windSpeed;
    float THI = (IS_METRIC ? 26 : 78);  // Heat index threshold Usually around 80F. (27C)
    // Old starting point.
    drawLabelValue(i, "Temperature:", String(currentWeather.temp, 1) + (IS_METRIC ? "°C" : "°F")); i++;
    if (T >= THI && heatIndex > T || 0) {  // Is heat index meaningful at colder temps? (Not really)
        drawLabelValue(i, "Heat Index:", String(heatIndex,1) + (IS_METRIC ? "°C" : "°F")); i++;
    }
    if (T <= TWC && windChill < T) {   
    //} else if (T < TWC) {   
        drawLabelValue(i, "Wind Chill:", String(windChill,1) + (IS_METRIC ? "°C" : "°F")); i++;
    } else if (T <= TWC && windChill >= T) {
      drawLabelValue(i, "Wind Chill:", "--"); i++;
    }
    drawLabelValue(i, "Humidity:", String(currentWeather.humidity) + "%"); i++;
    if (currentWeather.humidity >= 0) {  
        drawLabelValue(i, "Dew Point:", String((IS_METRIC ? dewPointC : dewPointF), 1) + (IS_METRIC ? "°C" : "°F")); i++; 
        //#ifdef NEVER
        // https://www.weather.gov/arx/why_dewpoint_vs_humidity

        drawLabelValue(i,"Feels like it is", feelsLike(TF, currentWeather.humidity, true) ); i++;           
        //#endif
    }
    drawLabelValue(i, "Wind Speed:", String(currentWeather.windSpeed, 1) + (IS_METRIC ? "m/s" : "mph") ); i++;
    drawLabelValue(i, "Wind Dir:", String(windDir(currentWeather.windDeg))); i++;
    drawLabelValue(i, "Pressure:", String(currentWeather.pressure) + "hPa"); i++;
    drawLabelValue(i, "Clouds:", String(currentWeather.clouds) + "%"); i++;
    if (IS_METRIC ) { // Weather "Visibility" has a rather sketchy definition.  (The US FAA doesn't acknowledge weather visibility)
      drawLabelValue(i, "Visibility:", String(currentWeather.visibility) + "m"); i++;
    } else {
      //drawLabelValue(i, "Visibility:", String(currentWeather.visibility) + "m"); i++;
      drawLabelValue(i, "Visibility:", String(fixMetricMiles(currentWeather.visibility),2) + " mi."); i++;
    }

    if (6 == 9) {
      String stringi = String(i);
      //drawLabelValue(i, currentWeather.description, String(i) );
      drawLabelValue(i, "i:", stringi ); i++;
      stringi = String(i);
      drawLabelValue(i, "i:", stringi ); i++;
    }    
  //}  
    //drawLabelValue(i, "Heap Mem:", String(ESP.getFreeHeap()) + " bytes"); i++;
  
    // Max of 4 lines if SHOWMOON is defined
    //drawLabelValue(i, "Wind chill", String(calcWindChill(T,V),1)); i++;
    //drawLabelValue(i, "Wind chill simple", String(windChillSimple(T,V),1)); i++;
   
    //drawLabelValue(i, "Flash Mem:", String(ESP.getFlashChipRealSize() / 1024 / 1024) + "MB"); i++;

  // String weatherIcon;
  // String weatherText;
}

// Needs to be #ifndeffed when running in TOUCH mode
// Assumes that we are on screen 0 and have limited space to display current conditions
// Setting the stage for more flexible screen 0.  i.e., !TOUCH and !GOTDHT is sort of supported now.
void shortDrawCurrentWeatherDetail() {
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  gfx.setColor(MINI_WHITE);
  // (Frogymandias mod)

  //float T = currentWeather.temp;
  //float TF = (IS_METRIC ? (T * 1.8) + 32 : T);
  //float V = currentWeather.windSpeed;
  float THI = (IS_METRIC ? 26 : 78);  // Heat index threshold Usually around 80F. (27C)
  //float TWC = (IS_METRIC ? TWCC : TWCF);   // Wind chill threshold Usually around 32F. (Maybe?)
  int i = 0;

  // Assumes that WS is mph.  Needs IS_METRIC fix
  // We are on screen 0 and are fitting things in between things that are already drawn.

        //drawLabelValue(i, "Out Dewpoint:", String((IS_METRIC ? dewPointC : dewPointF), 1) + (IS_METRIC ? "°C" : "°F")); i++; 
        // https://www.weather.gov/arx/why_dewpoint_vs_humidity
        //drawLabelValue(i,"Out Feels like:", feelsLike(TF, currentWeather.humidity, true) ); i++;
        //gfx.drawString(15, y + i*15,feelsLike(TF, currentWeather.humidity, false)); i++;
        //String outDP = "DP: " + String((IS_METRIC ? dewPointC : dewPointF), 1) + (IS_METRIC ? "°C" : "°F") + ". ";
        //outDP = outDP + feelsLike(TF, currentWeather.humidity, false);
        
        uint8_t y = 171;  // 210 Just below outside temp, humidity on screen 0
        gfx.setFont(ArialRoundedMTBold_14);
        gfx.setTextAlignment(TEXT_ALIGN_LEFT);
        gfx.setColor(MINI_YELLOW);
        gfx.drawString(4, y + i*15, "DP: ");
        gfx.setColor(MINI_WHITE);
        gfx.drawString(30, y + i*15, String((IS_METRIC ? dewPointC : dewPointF), 1) + (IS_METRIC ? "°C" : "°F"));
        gfx.setColor(MINI_YELLOW);
        gfx.drawString(77, y + i*15, feelsLike(TF, currentWeather.humidity, false)); i++;


        if (badDhtTemp == false && badHumidity == false) {
          float TdpF = calcDewpoint(dhtTempF, dhtHumidity); // calcDewpoint() always returns F
          float TdpC = (TdpF - 32) / 1.8;
          //drawLabelValue(i, "In  Dewpoint:", String( (IS_METRIC ? TdpC : TdpF), 1) + (IS_METRIC ? "°C" : "°F")); i++;
          // drawLabelValue(i, "In  Feels like:", feelsLike(dhtTempF, dhtHumidity, false) ); i++; 
          //String inDP = "DP: " + String( (IS_METRIC ? TdpC : TdpF), 1) + (IS_METRIC ? "°C" : "°F") + ". ";
          //inDP = inDP + feelsLike(dhtTempF, dhtHumidity, false);
          //gfx.drawString(4, y + i*15, inDP); i++;
          //gfx.setTextAlignment(TEXT_ALIGN_CENTER);
          //gfx.setColor(MINI_YELLOW);
          //gfx.drawString(15, y + i*15,feelsLike(dhtTempF, dhtHumidity, false)); i++;
          
          y=206;  // Just below inside temp, humidity on screen 0
          gfx.setFont(ArialRoundedMTBold_14);
          gfx.setTextAlignment(TEXT_ALIGN_LEFT);
          gfx.setColor(MINI_YELLOW);
          gfx.drawString(4, y + i*15, "DP: ");
          gfx.setColor(MINI_WHITE);
          gfx.drawString(30, y + i*15, String( (IS_METRIC ? TdpC : TdpF), 1) + (IS_METRIC ? "°C" : "°F"));
          gfx.setColor(MINI_YELLOW);
          gfx.drawString(77, y + i*15, feelsLike(dhtTempF, dhtHumidity, false)); i++;     
        }
        
    if ( currentWeather.windSpeed < 25 && !gotDHT) { 
      drawLabelValue(i, "Clouds:", String(currentWeather.clouds) + "%"); i++;
      drawLabelValue(i, "Pressure:", String(currentWeather.pressure) + "hPa"); i++;
      //drawLabelValue(i, "Wind Speed:", String(currentWeather.windSpeed, 1) + (IS_METRIC ? "m/s" : "mph") ); i++;
      //drawLabelValue(i, "Wind Dir:", String(windDir(currentWeather.windDeg))); i++;
    }
    if ( currentWeather.windSpeed >= 25) { // Assumes gotDHT
      drawLabelValue(i, "Wind Speed:", String(currentWeather.windSpeed, 1) + (IS_METRIC ? "m/s" : "mph") ); i++;
      drawLabelValue(i, "Wind Dir:", String(windDir(currentWeather.windDeg))); i++;
    } else {
      //drawLabelValue(i, "Clouds:", String(currentWeather.clouds) + "%"); i++;
      drawLabelValue(i, "Wind Speed:", String(currentWeather.windSpeed, 1) + (IS_METRIC ? "m/s" : "mph") ); i++;
      if (T >= THI && heatIndex > T) {  // What should we draw on the 4th line.
        drawLabelValue(i, "Heat index", String(heatIndex,1)); i++;
      } else if (T <=TWC && windChill < T) {   // Not sure what temp triggers wind chill being appropriate (Probably 50F)
      //} else if (T <= TWC) {   // 50F or 10C
        drawLabelValue(i, "Wind Chill", String(windChill,1) + (IS_METRIC ? "°C" : "°F")); i++;
      } else {
        if (T <= TWC) {                                  // Cold, but no wind chill
          drawLabelValue(i, "Wind Chill", "--"); i++;   //  Low wind == no wind chill
        } else { // Not cold
          //drawLabelValue(i, "Pressure:", String(currentWeather.pressure) + "hPa"); i++;
          drawLabelValue(i, "Wind Dir:", String(windDir(currentWeather.windDeg))); i++;
        }
        //drawLabelValue(i,"T: "+ String(T,1) + " TWC: " + String(TWC,1), "WC: " + String(windChill,1));  //  Debugging!!!
      }
    }

// Now on the 5th line
    if (isScreen0 && !gotDHT) { // We are on screen 0 but have never had a valid DHT reading
      if (IS_METRIC ) { // Weather "Visibility" has a rather sketchy definition.  (The US FAA doesn't acknowledge weather visibility)
        drawLabelValue(i, "Visibility:", String(currentWeather.visibility) + "m"); i++;
      } else {
        //drawLabelValue(i, "Visibility:", String(currentWeather.visibility) + "m"); i++;
        drawLabelValue(i, "Visibility:", String(fixMetricMiles(currentWeather.visibility),2) + " mi."); i++;
      }
    } // if (isScreen0 && !gotDHT)
  
    // Max of 4 lines if SHOWMOON is defined.  One or two more if !gotDHT
    // Things we mght want to see on screen 0, sometimes.
    // Temporarily comment out above lines to make room.
    //drawLabelValue(i, "Wind chill", String(calcWindChill(T,V),1)); i++;
    //drawLabelValue(i, "Wind chill simple", String(windChillSimple(T,V),1)); i++;
   
    //drawLabelValue(i, "Heap Mem:", String(ESP.getFreeHeap()) + " bytes"); i++;
    //drawLabelValue(i, "Flash Mem:", String(ESP.getFlashChipRealSize() / 1024 / 1024) + "MB"); i++;

  // String weatherIcon;
  // String weatherText;
}

void drawLabelValue(uint8_t line, String label, String value) {

  const uint8_t labelX = 10;  // Was 15
  const uint8_t valueX = 140; // (Frogymandias mod) Originally was 150, 146
  uint8_t y;
// (Frogymandias mod)
if (!isScreen0) { // (!TOUCH) screen 0 only can draw 4 lines.  Near the bottom.
  if (screen == 1) {  // Current conditions screen (are getting fragile yet?)
     y = 2;
  } else {  // Hopefully is screen 4 (About screen)
    y = 30; // Allows room for logo
  }
} else {
  if (gotDHT) { // On screen 0
    y = 210;
  } else {
    y = 207 - (2*15);
  }
} // if (!isScreen0)

  gfx.setTextAlignment(TEXT_ALIGN_LEFT);
  gfx.setColor(MINI_YELLOW);
  gfx.drawString(labelX, y + line * 15, label);
  gfx.setColor(MINI_WHITE);
  gfx.drawString(valueX, y + line * 15, value);
}

// converts the dBm to a range between 0 and 100%
int8_t getWifiQuality() {
  int32_t dbm = WiFi.RSSI();
  if (dbm <= -100) {
    return 0;
  } else if (dbm >= -50) {
    return 100;
  } else {
    return 2 * (dbm + 100);
  }
}

void drawWifiQuality() {
  int8_t quality = getWifiQuality();
  gfx.setColor(MINI_WHITE);
  gfx.setFont(ArialMT_Plain_10);
  gfx.setTextAlignment(TEXT_ALIGN_RIGHT);
  gfx.drawString(228, 9, String(quality) + "%");
  for (int8_t i = 0; i < 4; i++) {
    for (int8_t j = 0; j < 2 * (i + 1); j++) {
      if (quality > i * 25 || j == 0) {
        gfx.setPixel(230 + 2 * i, 18 - j);
      }
    }
  }
}

void drawForecastTable(uint8_t start) {
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  gfx.setColor(MINI_WHITE);
  gfx.drawString(120, 2, "Forecasts");
  uint16_t y = 0;

  String degreeSign = "°F";
  if (IS_METRIC) {
    degreeSign = "°C";
  }
  int firstColumnLabelX = 50;
  int firstColumnValueX = 70;
  int secondColumnLabelX = 130;
  int secondColumnValueX = 170;
  for (uint8_t i = start; i < start + 4; i++) {
    gfx.setTextAlignment(TEXT_ALIGN_LEFT);
    y = 45 + (i - start) * 75;
    if (y > 320) {
      break;
    }
    gfx.setColor(MINI_WHITE);
    gfx.setTextAlignment(TEXT_ALIGN_CENTER);
    time_t time = forecasts[i].observationTime;
    struct tm * timeinfo = localtime (&time);
    if (IS_STYLE_12HR) {
      gfx.drawString(120, y - 15, WDAY_NAMES[timeinfo->tm_wday] + " " + String(make12_24(timeinfo->tm_hour)));
    } else {
      gfx.drawString(120, y - 15, WDAY_NAMES[timeinfo->tm_wday] + " " + String(timeinfo->tm_hour) + ":00");
    }

    gfx.drawPalettedBitmapFromPgm(0, 5 + y, getMiniMeteoconIconFromProgmem(forecasts[i].icon));
    gfx.setTextAlignment(TEXT_ALIGN_LEFT);
    gfx.setColor(MINI_YELLOW);
    gfx.setFont(ArialRoundedMTBold_14);
    gfx.drawString(0, y - 8, forecasts[i].main);
    gfx.setTextAlignment(TEXT_ALIGN_LEFT);

    gfx.setColor(MINI_BLUE);
    gfx.drawString(firstColumnLabelX, y, "T:");
    gfx.setColor(MINI_WHITE);
    gfx.drawString(firstColumnValueX, y, String(forecasts[i].temp, 0) + degreeSign);

    gfx.setColor(MINI_BLUE);
    gfx.drawString(firstColumnLabelX, y + 15, "H:");
    gfx.setColor(MINI_WHITE);
    gfx.drawString(firstColumnValueX, y + 15, String(forecasts[i].humidity) + "%");

    gfx.setColor(MINI_BLUE);
    gfx.drawString(firstColumnLabelX, y + 30, "P: ");
    gfx.setColor(MINI_WHITE);
    gfx.drawString(firstColumnValueX, y + 30, String(fixMetricRain(forecasts[i].rain), 2) + (IS_METRIC ? "mm" : "in"));

    gfx.setColor(MINI_BLUE);
    gfx.drawString(secondColumnLabelX, y, "Pr:");
    gfx.setColor(MINI_WHITE);
    gfx.drawString(secondColumnValueX, y, String(forecasts[i].pressure, 0) + "hPa");

    gfx.setColor(MINI_BLUE);
    gfx.drawString(secondColumnLabelX, y + 15, "WSp:");
    gfx.setColor(MINI_WHITE);
    gfx.drawString(secondColumnValueX, y + 15, String(forecasts[i].windSpeed, 0) + (IS_METRIC ? "m/s" : "mph") );

    gfx.setColor(MINI_BLUE);
    gfx.drawString(secondColumnLabelX, y + 30, "WDi: ");
    gfx.setColor(MINI_WHITE);
    gfx.drawString(secondColumnValueX, y + 30, String(windDir(forecasts[i].windDeg)));
  }
}

void drawAbout() {
  gfx.fillBuffer(MINI_BLACK);
  gfx.drawPalettedBitmapFromPgm(20, 5, ThingPulseLogo);

  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  gfx.setColor(MINI_WHITE);
  gfx.drawString(120, 90, "https://thingpulse.com");

  bool Sleepable = false;
  if (Sleep_Interval && SLEEP_INTERVAL_SECS) Sleepable = true;

// See https://arduino-esp8266.readthedocs.io/en/latest/libraries.html#esp-specific-apis

  // i++; is a Frogymandias mod
  int i = 6;  // to account for the ThingPulse logo at the top of the screen
  gfx.setFont(ArialRoundedMTBold_14);
  gfx.setTextAlignment(TEXT_ALIGN_CENTER);
  drawLabelValue(i, "Version:", VERSION); i++;
  drawLabelValue(i, "Heap Mem:", String(ESP.getFreeHeap() / 1024) + "kb"); i++;
  //drawLabelValue(i, "Heap Mem:", String(ESP.getFreeHeap()) + " bytes"); i++;
  drawLabelValue(i, "Heap Fragment:", String(ESP.getHeapFragmentation()) + "%"); i++;  // (0% is clean, more than ~50% is not harmless)
  if (!Sleepable) {drawLabelValue(i, "Flash Mem:", String(ESP.getFlashChipRealSize() / 1024 / 1024) + "MB"); i++;}
  //drawLabelValue(i, "Sketch size:", String(ESP.getSketchSize() / 1024) + "kb"); i++;
  //drawLabelValue(i, "Free sketch:", String(ESP.getFreeSketchSpace() / 1024) + "kb"); i++;

  drawLabelValue(i, "WiFi Strength:", String(WiFi.RSSI()) + "dB"); i++;
  drawLabelValue(i, "Chip ID:", String(ESP.getChipId())); i++;
  //Serial.printf("Chip ID: ");
  //String ID = String(ESP.getChipId() );
  //Serial.printf(String(ESP.getChipId() ) ); // can't be done.
  //Serial.printf (ID);
  //Serial.printf("\n");

  drawLabelValue(i, "VCC: ", String(ESP.getVcc() / 1024.0) + "V"); i++; // Requires ADC_MODE(ADC_VCC), which uses the analog in/output pin.
  //drawLabelValue(i, "CPU Freq.: ", String(ESP.getCpuFreqMHz()) + "MHz"); i++;

  char time_str[15];
  const uint32_t millis_in_day = 1000 * 60 * 60 * 24;
  const uint32_t millis_in_hour = 1000 * 60 * 60;
  const uint32_t millis_in_minute = 1000 * 60;
  uint8_t days = millis() / (millis_in_day);
  uint8_t hours = (millis() - (days * millis_in_day)) / millis_in_hour;
  uint8_t minutes = (millis() - (days * millis_in_day) - (hours * millis_in_hour)) / millis_in_minute;
  sprintf(time_str, "%2dd%2dh%2dm", days, hours, minutes);
  
  drawLabelValue(i, "Uptime: ", time_str); i++;
  uint16_t remainingSecs = UPDATE_INTERVAL_SECS-(millis() - lastDownloadUpdate)/1000;
  if (Sleepable) {
    int remainingSleep = SLEEP_INTERVAL_SECS-((millis() - timerPress)/1000);
    drawLabelValue(i, "Will sleep in: ", String(remainingSleep) + " secs."); i++;
   }
  drawLabelValue(i, "Next update in: ", String(remainingSecs) + " secs."); i++;
  drawLabelValue(i, "IP Address: ", WiFi.localIP().toString()); i++;
  
  gfx.setTextAlignment(TEXT_ALIGN_LEFT);
  gfx.setColor(MINI_YELLOW);
  gfx.drawString(15, 280, "Last Reset: ");
  gfx.setColor(MINI_WHITE);
  gfx.drawStringMaxWidth(15, 295, 240 - 2 * 15, ESP.getResetInfo());
}

void calibrationCallback(int16_t x, int16_t y) {
  gfx.setColor(1);
  gfx.fillCircle(x, y, 10);
}

String getTime(time_t *timestamp) {
  struct tm *timeInfo = localtime(timestamp);

  char buf[9];  // "12:34 pm\0"
  char ampm[3];
  ampm[0]='\0'; //Ready for 24hr clock
  uint8_t hour = timeInfo->tm_hour;

  if (IS_STYLE_12HR) {
    if (hour > 12) {
      hour = hour - 12;
      sprintf(ampm, "pm");
    } else {
      sprintf(ampm, "am");
    }
    sprintf(buf, "%2d:%02d %s", hour, timeInfo->tm_min, ampm);
  } else {
    sprintf(buf, "%02d:%02d %s", hour, timeInfo->tm_min, ampm);
  }
  return String(buf);
}

/*
 *  Convert hour from 24 hr time to 12 hr time.
 *
 *  @return cString with 2 digit hour + am or pm 
 */
char* make12_24(int hour){
  static char hr[6];
  if(hour > 12){
    sprintf(hr, "%2d pm", (hour -12) );
    //sprintf(buf, "%02d:%02d %s", hour, timeInfo->tm_min, ampm);
  } else {
    sprintf(hr, "%2d am", hour);
  }
  return hr;
}

// (Frogymandias mod)   
#ifdef TOUCH
  void loadPropertiesFromSpiffs() {
    if (SPIFFS.begin()) {
      const char *msg = "Using '%s' from SPIFFS\n";
      Serial.println("Attempting to read application.properties file from SPIFFS.");
      File f = SPIFFS.open("/application.properties", "r");
      if (f) {
        Serial.println("File exists. Reading and assigning properties.");
        while (f.available()) {
          String key = f.readStringUntil('=');
          String value = f.readStringUntil('\n');
          if (key == "ssid") {
            WIFI_SSID = value.c_str();
            Serial.printf(msg, "ssid");
          } else if (key == "password") {
            WIFI_PASS = value.c_str();
            Serial.printf(msg, "password");
          } else if (key == "timezone") {
            TIMEZONE = getTzInfo(value.c_str());
            Serial.printf(msg, "timezone");
          } else if (key == "owmApiKey") {
            OPEN_WEATHER_MAP_API_KEY = value.c_str();
            Serial.printf(msg, "owmApiKey");
          } else if (key == "owmLocationId") {
            OPEN_WEATHER_MAP_LOCATION_ID = value.c_str();
            Serial.printf(msg, "owmLocationId");
          } else if (key == "locationName") {
            DISPLAYED_LOCATION_NAME = value.c_str();
            Serial.printf(msg, "locationName");
          } else if (key == "isMetric") {
            IS_METRIC = value == "true" ? true : false;
            Serial.printf(msg, "isMetric");  
          } else if (key == "is12hStyle") {
            IS_STYLE_12HR = value == "true" ? true : false;
            Serial.printf(msg, "is12hStyle");
          }
        }
      } else {
        Serial.println("Does not exist.");
      }
      f.close();
      Serial.println("Effective properties now as follows:");
      Serial.println("\tssid: " + WIFI_SSID);
      Serial.println("\tpassword: " + WIFI_PASS);
      Serial.println("\ttimezone: " + TIMEZONE);
      Serial.println("\tOWM API key: " + OPEN_WEATHER_MAP_API_KEY);
      Serial.println("\tOWM location id: " + OPEN_WEATHER_MAP_LOCATION_ID);
      Serial.println("\tlocation name: " + DISPLAYED_LOCATION_NAME);
      Serial.println("\tmetric: " + String(IS_METRIC ? "true" : "false"));
      Serial.println("\t12h style: " + String(IS_STYLE_12HR ? "true" : "false"));
    } else {
      Serial.println("SPIFFS mount failed.");
    }
  }
#endif

// (Frogymandias mod)
#ifdef TOUCH
/*
 * Change screen based on touchpoint location.
 */
uint8_t changeScreen(TS_Point p, uint8_t screen) {
  uint8_t page = screen;

#define VERBOSE_TOUCHPOINTS
#ifdef VERBOSE_TOUCHPOINTS
  Serial.printf("Touch point detected at %d/%d.", p.x, p.y);
  // From the screen's point of view commented values for the 240 X 320 touch screen
  if (p.y < dividerTop)      Serial.print(" top ");    // < 80
  if (p.y > dividerBottom)   Serial.print(" bottom "); // > 240
  if (p.x > dividerMiddle)   Serial.print(" left ");   // > 120
  if (p.x <= dividerMiddle)  Serial.print(" right ");  // <= 120
#endif

  if (p.y < dividerTop) {            // top -> change 12/24h style
    // IS_STYLE_12HR = !IS_STYLE_12HR;
    if (page == 0) { // Only toggle if pressed on screen 0. 
      if (TOGGLE_12HR) { // (Frogymandias) See settings.h
        IS_STYLE_12HR = !IS_STYLE_12HR;
        #ifdef VERBOSE_TOUCHPOINTS
        Serial.printf("Toggle to %s mode ", (IS_STYLE_12HR) ? "AM/PM" : "24 hour");
        #endif
      } else if (SLEEP_INTERVAL_SECS && TOGGLE_SLEEP) {  // (Frogymandias) Toggle sleep mode on/off
        Sleep_Interval = !Sleep_Interval;
        #ifdef VERBOSE_TOUCHPOINTS
        Serial.printf("Toggle to %s mode ", (Sleep_Interval) ? "Sleep" : "Not sleep");
        #endif
      }
    } // if (page == 0) 
  } else if (p.y > dividerBottom) {  // bottom -> go to screen 0
    page = 0;
  } else if (p.x > dividerMiddle) {  // left -> previous page
    if (page == 0) {            // Note type is unsigned
      page = screenCount;       // Last screen is max -1
    }
    page--;
  } else {                      // right -> next screen
    page = (page + 1) % screenCount;
  }
// (Frogymandias mod)
#ifdef VERBOSE_TOUCHPOINTS
  Serial.printf(" page = %d\n", page);
  Serial.println();
#endif
  return page;
}
#endif  // #ifdef TOUCH

// (Frogymandias mod)  Assumes that TOUCH is not defined.
#ifdef BUTTON_SWITCHES

uint8_t readButtons(uint8_t screen) {
  uint8_t but1 = readButton1(screen);
  uint8_t but2 = readButton2(screen);
  return (but1 | but2);
 }
 
uint8_t readButton1(uint8_t screen) {

bool shortPress = false;
bool longPress = false;
uint32_t veryLongTime = 0;  // To allow for very long presses. 

uint8_t screenMem = screen;
bool temp = isScreen0;

  int debounce;
  if (screen == 0) {
    debounce = BUTTON_DEBOUNCE/4;   // Screen 0 is busy.
  } else {
    debounce = BUTTON_DEBOUNCE;
  }

  int switchStatus1 = digitalRead(BUTTON_1); 
  if (switchStatus1 == LOW) {
    delay(debounce); // Debounce
    switchStatus1 = digitalRead(BUTTON_1);
  }
  // long press logic;
  if (switchStatus1 == LOW) {
    veryLongTime = veryLongTime + debounce;
    shortPress = true;   // At this point we at least have a short press
    delay(BUTTON_LONG_DELAY);
    veryLongTime = veryLongTime + BUTTON_LONG_DELAY;
    switchStatus1 = digitalRead(BUTTON_1);
    while (switchStatus1 == LOW) {  // We are in a long press.  Let's ride it out.
      if ( (6 == 9)  ) { // Maybe is too confusing to change screens during a longpress.
        isScreen0 = false;
        drawCurrentWeatherDetail();
        gfx.commit();
        Serial.printf("drawCurrentWeatherDetail()\n");
      }
      longPress = true;
      delay(BUTTON_LONG_DELAY);
      veryLongTime = veryLongTime + BUTTON_LONG_DELAY;
      switchStatus1 = digitalRead(BUTTON_1);

    }
  } // if (switchStatus1 == LOW) If false we have no presses
  screen = screenMem;
  isScreen0 = temp;
  if (veryLongTime >= VERY_LONG_TIME * 1000) {
    Serial.printf("Button 1 had a very long press (%d milliseonds)\n", veryLongTime);
    return BUTTON_1_VERY_LONG;
  } else if (longPress == true) {
    Serial.printf("Button 1 had a long press. (%d milliseonds)\n", veryLongTime);
    return(BUTTON_1_LONG); 
  } else if (shortPress == true) {
    veryLongTime = veryLongTime - BUTTON_LONG_DELAY;
    Serial.printf("Button 1 had a short press. (%d milliseonds)\n", veryLongTime);
    return(BUTTON_1_SHORT);
  } else {
    return 0; // Button 1 wasn't pressed.  At least not long enough for 1 debounce delay.
  }
}

uint8_t readButton2(uint8_t screen) {

bool shortPress = false;
bool longPress = false;
long longTime = 0;
uint32_t veryLongTime = 0;

uint8_t screenMem = screen;
bool temp = isScreen0;

  int debounce;
  if (screen == 0) {
    debounce = BUTTON_DEBOUNCE/4; 
  } else {
    debounce = BUTTON_DEBOUNCE;     
  }

  int switchStatus2 = digitalRead(BUTTON_2); 
  if (switchStatus2 == LOW) {
    delay(debounce); // Debounce
    switchStatus2 = digitalRead(BUTTON_2);
  }
  // long press logic;
  if (switchStatus2 == LOW) { // We need to keep it pressed at least debounce time to register as something.
    veryLongTime = veryLongTime + debounce;
    shortPress = true;   // At this point we at least have a short press
    delay(BUTTON_LONG_DELAY);
    veryLongTime = veryLongTime + BUTTON_LONG_DELAY; // We always have to wait through the delay, so it gets counted.
    switchStatus2 = digitalRead(BUTTON_2);
    while (switchStatus2 == LOW) {  // We are in a long press.  Let's ride it out.
      if ( (6 == 9) && (SLEEP_INTERVAL_SECS == 0) && (TOGGLE_12HR == false)) {  // Maybe is too confusing to change screens during a long press.
        isScreen0 = false;
        drawCurrentWeatherDetail();
        gfx.commit();
        Serial.printf("drawCurrentWeatherDetail()\n");
      }
      longPress = true;
      delay(BUTTON_LONG_DELAY);
      switchStatus2 = digitalRead(BUTTON_2);
      veryLongTime = veryLongTime + BUTTON_LONG_DELAY;
    }
  } // if (switchStatus2 == LOW) If false we have no presses
  screen = screenMem;
  isScreen0 = temp;
  if (veryLongTime >= VERY_LONG_TIME * 1000) { // Say, at least 5 seconds
    Serial.printf("Button 2 had a very long press (%d milliseonds)\n", veryLongTime);
    return BUTTON_2_VERY_LONG;
  } else if (longPress == true) {
    Serial.printf("Button 2 had a long press. (%d milliseonds)\n", veryLongTime);
    return(BUTTON_2_LONG); 
  } else if (shortPress == true) {
     veryLongTime = veryLongTime - BUTTON_LONG_DELAY;
    Serial.printf("Button 2 had a short press. (%d milliseconds)\n", veryLongTime);
    return(BUTTON_2_SHORT);
  } else {
    return 0; // Button 2 wasn't pressed.
  }
}

uint8_t changeButtonScreen(uint8_t buttonState, uint8_t screen) {
  uint8_t page = screen;
 
  if ( buttonState == BUTTON_1_SHORT) {
    if (page == 0) {            // Note type is unsigned
      page = screenCount;       // Last screen is max -1
    }
    page--;
    Serial.print(" left (BUTTON_1 D4)");
    Serial.printf(" page = %d\n", page);
    //Serial.println();
  } else if (buttonState == BUTTON_2_SHORT ) {
    page = (page + 1) % screenCount;
    if (page == 1 && 0) {  // Probably need some kind of option here.  But settings.h is already big enough.
      Serial.print(" Skipping page 1 ");  // No touch has most current data on screen 0 so screen 1 is (mostly) redundant.
      page = (page + 1) % screenCount;
    }
    Serial.print(" right (BUTTON_2 D3 (Flash))");
    Serial.printf(" page = %d\n", page);
    //Serial.println();
  }
  return page;
}
#endif  // #ifdef BUTTON_SWITCHES


#ifndef NEVER // Remove this #ifdef--we aren't running out of memory
// US NWS The Heat Index Equation (in F)
// https://www.wpc.ncep.noaa.gov/html/heatindex_equation.shtml
float calcHeatIndex(float T, float RH) {
  
  // First do the simple calculation for HI
  float SF = 0.5 * (T + 61.0 + ((T-68.0)*1.2) + (RH*0.094)) ;
  // Then do this calculation to decide if we need to use the complex formula
  float avg = ((SF + T) / 2);
  
  float HI;
  float Thresh = 80;  // 80 degrees F for real use
  if (avg < Thresh) {
    HI = SF;
    Serial.println("(avg) < 80F.  Using simple HI formula.");
  } else {
    Serial.println("(avg) >= 80F.  Using complex HI formula.");
    HI = -42.379 + 2.04901523*T + 10.14333127*RH - .22475541*T*RH - .00683783*T*T - .05481717*RH*
    RH + .00122874*T*T*RH + .00085282*T*RH*RH - .00000199*T*T*RH*RH;
  
    // And maybe need adjustments...
    float adj;
    if ( (RH < 13) && ( (T >= 80) && (T <= 112) ) ) {
      adj = ((13-RH)/4) * sqrt((17-abs(T-95.))/17);
      HI = HI - adj;
      Serial.println("Doing HI adjustment 1");
    }
  
    if ((RH > 85) && ( (T >= 80) && (T <= 87) ) ) {    
      adj = ((RH-85)/10) * ((87-T)/5);
      HI = HI + adj;
      Serial.println("Doing adjustment 2: ADJUSTMENT = [(RH-85)/10] * [(87-T)/5]");
      //Serial.print("Adjustment HI 2");
    }
  }
    Serial.println("Temp: " + String(T,1) + "F Rel humidity: " + String(RH,0) + "% Heat index: " + String (HI,1));
    return HI;
} // calcHeatIndex()
# else
// US NWS The Heat Index Equation (in F)
// https://www.wpc.ncep.noaa.gov/html/heatindex_equation.shtml
float calcHeatIndex(float T, float RH) {
  
  // First do the simple calculation for HI
  // And that's it.
  return 0.5 * (T + 61.0 + ((T-68.0)*1.2) + (RH*0.094)) ;
}
#endif

// The New Improved Wind Chill Index
// https://www.weather.gov/ffc/wci
//
//                                           ummmm?
// Wind Chill (oF) = 35.74 + 0.6215T - 35.75(V0.16) + 0.4275T(V0.16)
// Where, T = Air Temperature (oF)
// V = Wind Speed (mph)
float calcWindChill(float T, float V) {
   // https://www.backpacker.com/skills/how-to-calculate-wind-chill/
   float WC = 35.74 + (0.6215*T) - 35.75*(pow(V,0.16)) + (0.4275*T*(pow(V,0.16)));
   if (T > 50 || V < 3.0) {
      Serial.println("Can't calclate windchill so forcing it to T");
      //Serial.println( (((T>50) ? "TF > 50 " : " ") + ((V < 3.0) ? "Wind speed < 3.0" : "") ); 
      if ( T > 50) {
        Serial.println("TF > 50F "); 
      } else if(V < 3.0) {
        Serial.println("Wind speedF < 3.0mph");
      } 
      WC = T;
   }
   if (WC > T ) {
          Serial.println("Wind chill is too high (will fix): " + String(WC,1) + "°F");
          WC = T;
   }
   //     Serial.println("Temp: " + String(T,1) + "F Rel humidity: " + String(RH,0) + "% Heat index: " + String (HI,1));
   Serial.println("Temp: " + String(T,1) + "°F windSpeed: " + String(V,1) + "mph"+ " Wind chill: " + String(WC,1) + "°F");
   return WC;
  //return 35.74 + (0.6215*T) - 35.75*(pow(V,0.16)) + (0.4275*T*(pow(V,0.16)));
}

#ifdef NEVER
float windChillSimple (float T, float V) {

  // A good approximation of the wind-chill temperature can be found by multiplying the 
  // wind speed by 0.7 and then subtracting that value from the air temperature.
  // https://www.chicagotribune.com/weather/ct-wea-asktom-0120-20180119-column.html 

  return T - (V * .7);
}
#endif

#ifndef NEVER
// Converts wind direction in degrees to NSEW notation
// http://snowfence.umn.edu/Components/winddirectionanddegrees.htm
String windDir(float dir) {  // 16 divisions
  if ((dir >= 348.75) && (dir <= 360) || (dir >= 0) && (dir <= 11.25)) {
    return "N";
  } else if ((dir >= 11.25) && (dir <= 33.75)) {
    return "NNE";
  } else if ((dir >= 33.45) && (dir <= 56.25)) {
    return "NE";
  } else if ((dir >= 56.25) && (dir <= 78.75)) {
    return "ENE";
  } else if ((dir >= 78.75) && (dir <= 101.25)) {
    return "E";
  } else if ((dir >= 101.25) && (dir <=123.75 )) {
    return "ESE";
  } else if ((dir >= 123.75) && (dir <= 146.25)) {
    return "SE";
  } else if ((dir >= 146.25) && (dir <= 168.75)) {
    return "SSE";
  } else if ((dir >= 168.75) && (dir <= 191.25)) {
    return "S";
  } else if ((dir >= 191.25) && (dir <= 213.75)) {
    return "SSW";
  } else if ((dir >= 213.75) && (dir <= 236.25)) {
    return "SW";
  } else if ((dir >= 236.25) && (dir <= 258.75)) {
    return "WSW";
  } else if ((dir >= 258.75) && (dir <= 281.25)) {
    return "W";
  } else if ((dir >= 281.25) && (dir <= 303.75)) {
    return "WNW";
  } else if ((dir >= 303.75) && (dir <= 326.25)) {
    return "NW";
  } else if ((dir >= 326.25) && (dir <= 348.75)) {
    return "NNW";
  } else {
    return String(dir,0) + "?";
  }
}
#else
String windDir(float dir) {  // 8 divisions version is only if we are really crampled for memory
  if ((dir >= 337.5) && (dir <= 360) || (dir >= 0) && (dir <= 22.25)) {
    return "N";
  } else if ((dir >= 22.5) && (dir <= 67.5)) {
    return "NE";
  } else if ((dir >= 67.5) && (dir <= 112.5)) {
    return "E";
  } else if ((dir >= 112.5) && (dir <= 157.5)) {
    return "SE";
  } else if ((dir >= 157.5) && (dir <= 202.5)) {
    return "S";
  } else if ((dir >= 202.5) && (dir <= 247.5)) {
    return "SW";
  } else if ((dir >= 247.5) && (dir <= 292.5)) {
    return "W";
  } else if ((dir >= 292.5) && (dir <= 337.5)) {
    return "NW";
  } else {
    return String(dir,0) + "?";
  }
}
#endif

// (Frogymandias mod) (This simple approximation was usually way off)
// https://en.wikipedia.org/wiki/Dew_point#Simple_approximation
// Tf is degrees Fahrenheit 
// Celsius is (Tf - 32) / 1.8
// This simple approximation is valid if RH > 50% (in some alternate universe)
float calcDewpointxx (float Tf, float RH) {
  float Tdp = Tf - (9/25 * (100-RH)); // Does not work. Always returns Tf.
  //float Tdp = Tf - (0.36 * (100-RH));     // Works.  Returns correct Tdp.
  //return (IS_METRIC ? (Tdp-32)/1.8 : Tdp); 
  return Tdp; // Always returns Tdp as F.
}

float calcDewpoint (float Tf, float RH) {
// Formula is in Celcius, so
//float Tc = (Tf - 32) / 1.8; //!!
float Tc = (Tf - 32) *.5556;  // 11/23/2022

// http://irtfweb.ifa.hawaii.edu/~tcs3/tcs3/Misc/Dewpoint_Calculation_Humidity_Sensor_E.pdf
float H = (log10(RH)-2)/0.4343 + (17.62*Tc)/(243.12+Tc);
float DpC = 243.12*H/(17.62-H); // Dew point in Celcius

float DpF = (DpC * 1.8) + 32; // Back to F
return DpF;

// DpF agrees with the interactive DP calculator at
// https://www.omnicalculator.com/physics/dew-point
}

#ifdef NEVER
// Feels like in the *summer* (This needs expanding to cover all temps/humidities, not just "outside in the summer")
// https://www.weather.gov/arx/why_dewpoint_vs_humidity
// "So if you want a real judge of just how "dry" or "humid" it will feel outside, look at the
// dew point instead of the RH. The higher the dew point, the muggier it will feel."
String feelsLikeNWS(float TF, float humidity, bool shortDisplay) {
  float dewPointF = calcDewpoint(TF, humidity);
  String feel = "";
  if (dewPointF < 41) {
    feel = (shortDisplay = true ? "Dry" : "Dry");
  } else if (dewPointF >= 41 && dewPointF <= 55) { 
    feel = (shortDisplay == true ? "Dry, Comfy" : "Dry and comfortable");
  } else if (dewPointF >= 55 && dewPointF <= 65) {
    feel = (shortDisplay == true ? "Sticky, Muggy" : "'Sticky' with muggy evenings"); // Too long.  Shorten
  } else if (dewPointF >= 65) {
    feel = (shortDisplay == true ? "Oppressive" :  "Oppressive. Be careful."); //"Lots of moisture in the air, becoming oppressive");
  } else {
    feel = "Dunno";
  }
  return feel;
}
#endif

// Omni Calclator criteria (very roughly)
// https://www.omnicalculator.com/physics/dew-point
String feelsLike(float TF, float humidity, bool shortDisplay) {
  float dewPointF = calcDewpoint(TF, humidity);
  //dewPointF = dewPointF + 20; // Only for testing!
  //String feel = "";
  // String feel; // Is now a global.
  feel.reserve(100);
  feel = "Dunno";
  if (humidity <= 20) {
    feel = (shortDisplay == true ? "Xtrmly dry" : "Extremely dry");
  } else if (humidity >= 20 && dewPointF <= 40) {  // !!
    feel = (shortDisplay == true ? "Very dry" : "Very dry");
  } else if (dewPointF >= 40 && dewPointF <= 48) {
    if (TF >= 70) {
      feel = (shortDisplay == true ? "Dryer" : "Warm but a bit dry");
    }
      feel = (shortDisplay == true ? "Dryer" : "A bit dry");
  } else if (dewPointF >= 48 && dewPointF <= 55) { 
    if (TF < 67) {
      feel = (shortDisplay == true ? "Dry" : "Dry, but chilly");
    } else if (TF >= 80) {
      feel = (shortDisplay == true ? "Dry, warm" : "Dry, but warm");
    } else {
      feel = (shortDisplay == true ? "Dry, Comfy" : "Dry and comfortable"); 
    }
  } else if (dewPointF >= 55 && dewPointF <= 60) { 
    feel = (shortDisplay == true ? "Moistish" : "Dryish but getting moist");  // Max characters
  } else if (dewPointF >= 60 && dewPointF <= 65) {
    feel = (shortDisplay == true ? "Getting sticky" : "Getting sticky");
  } else if (dewPointF >= 65 && dewPointF <= 70) { // "Dryish but getting moist"   
    feel = (shortDisplay == true ? "Unpleasant" : "Unpleasant, wet air");  //"Unpleasant, lots of moisture in the air");    
  } else if (dewPointF >= 70) {                // "Can be danger zone. ";
    if (TF <= 75) {  // But not sure about the "danger" warning. (Softened the danger warning.)
      feel = (shortDisplay == true ? "Oppressive" : "Oppressive.");//"Oppressive, even dangerous above 75 °F");
    } else {                                     // "Dryish but getting moist"
      feel = (shortDisplay == true ? "Oppressive" :  "Oppressive. Be careful.");// "Oppressive. Dangerous above 75 °F");
    }                                         //    "Can be danger zone." // T+DP. Can be danger.
                                              //    "Dryish but getting moist"
  } else {
    feel = "Dunno";
  }
  return feel;
}
